/* Simple collision detector for FX Calculators, by Loïc Payol */

#include "coll.h"
#include <math.h>
#include <stdlib.h>

static C_Vector LineToVector(const C_Line *l)
{
	C_Vector v;
	v.x =l->x2 - l->x1; v.y = l->y2 - l->y1;
	return v;
}
static C_Vector SegmentToVector(const C_Segment *s)
{
	C_Vector v;
	v.x = s->x2 - s->x1; v.y = s->y2 - s->y1;
	return v;
}

static C_Line SegmentToLine(const C_Segment *s)
{
	C_Line l;
	l.x1 = s->x1; l.y1 = s->y1;	l.x1 = s->x2; l.y2 = s->y2;
	return l;
}

static C_Bool ProjectSegment(const int Cx,const int Cy,const int Ax,const int Ay,const int Bx,const int By)
{
	int ACx = Cx-Ax;
	int ACy = Cy-Ay; 
	int ABx = Bx-Ax;
	int ABy = By-Ay; 
	int BCx = Cx-Bx;
	int BCy = Cy-By; 
	int s1 = (ACx*ABx) + (ACy*ABy);
	int s2 = (BCx*ABx) + (BCy*ABy); 
	if (s1*s2 > 0)
		return FALSE;
	return TRUE;
}

#ifdef COLLIDE_TWOAABB
C_Bool C_TwoAABB(const C_AABB *b1, const C_AABB *b2)
{
	return ((b2->x >= b1->x + b1->w)
		   || (b2->x + b2->w <= b1->x)
		   || (b2->y >= b1->y + b1->h)
		   || (b2->y + b2->h <= b1->y)) ? FALSE : TRUE;
}
#endif

#ifdef COLLIDE_POINTCIRCLE
C_Bool C_PointCircle(const C_Point *p, const C_Circle *c)
{
	if(SQUARE(p->x - c->x) + SQUARE(p->y - c->y) > SQUARE(c->radius))
		return FALSE;
	else
		return TRUE;
}
#endif

#ifdef COLLIDE_LINECIRCLE
C_Bool C_LineCircle(const C_Line *l, const C_Circle *c)
{
	C_Vector u = LineToVector(l), AC;
	double num, dem;
	AC.x = c->x - l->x1; AC.y = c->y - l->y1;
	num = u.x * AC.y - u.y * AC.x;
	if(num < 0)
		num = -num;
	dem = sqrt(u.x*u.x + u.y*u.y);
	if(num / dem < c->radius)
		return TRUE;
	return FALSE;
}
#endif

#ifdef COLLIDE_SEGMENTCIRCLE
C_Bool C_SegmentCircle(const C_Segment *s, const C_Circle *c)
{
	C_Vector AB = SegmentToVector(s), AC, BC;
	C_Line line = SegmentToLine(s);
	C_Point A, B;
	double pscal1 = AB.x*AC.x + AB.y*AC.y;
	double pscal2 = (-AB.x)*BC.x + (-AB.y)*BC.y;
	AC.x = c->x - s->x1;
	AC.y = c->y - s->y1;
	BC.x = c->x - s->x2;
	BC.y = c->y - s->y2;
	if(C_LineCircle(&line, c) == FALSE)
		return FALSE;
	if(pscal1 >= 0 && pscal2 >= 0)
		return TRUE;
	A.x = s->x1; A.y = s->y1;
	B.x = s->x2; B.y = s->y2;
	if(C_PointCircle(&A,c))
		return TRUE;
	if(C_PointCircle(&B, c))
		return TRUE;
	return FALSE;
}
#endif

#ifdef COLLIDE_POINTAABB
C_Bool C_PointAABB(const C_Point *p, const C_AABB *b)
{
	if(	p->x >= b->x
		&& p->x < b->x + b->w
		&& p->y >= b->y
		&& p->y < b->y + b->h )
		return TRUE;
	return FALSE;
}
#endif

#ifdef COLLIDE_CIRCLEAABB
C_Bool C_CircleAABB(const C_Circle *c, const C_AABB *b)
{
	C_AABB bc;
	C_Point d, e, f, g, center;
	bc.x = c->x - c->radius; bc.y = c->y - c->radius; bc.w = c->radius * 2; bc.h = bc.w;
	d.x = b->x; d.y = b->y;
	e.x = b->x; e.y = b->y + b->h;
	f.x = b->x + b->w; f.y = b->y + b->h;
	g.x = b->x + b->w; g.y = b->y;
	center.x = c->x; center.y = c->y;
	if(C_TwoAABB(b, &bc) == FALSE)
		return FALSE;
	if(
		C_PointCircle(&d, c)
		|| C_PointCircle(&e, c)
		|| C_PointCircle(&f, c)
		|| C_PointCircle(&g, c)
	)
		return TRUE;
	if(C_PointAABB(&center, b))
		return TRUE;
	if(ProjectSegment(c->x, c->y, b->x, b->y, b->x, b->y + b->h) || ProjectSegment(c->x, c->y, b->x, b->y, b->x + b->w, b->y))
		return TRUE;
	return FALSE;
}
#endif

#ifdef COLLIDE_TWOCIRCLE
C_Bool C_TwoCircle(const C_Circle *c1, const C_Circle *c2)
{
	if(SQUARE(c1->x - c2->x) + SQUARE(c1->y - c2->y) > SQUARE(c1->radius + c2->radius))
		return FALSE;
	return TRUE;
}
#endif

#ifdef COLLIDE_POINTTWOCIRCLE
C_Bool C_PointTwoCircle(C_Point *p, const C_Circle *c1, const C_Circle *c2)
{
	if(C_TwoCircle(c1, c2))	{
		p->x = ((c1->x * c2->radius) + (c2->x * c1->radius)) / (c1->radius + c2->radius);
		p->y = ((c1->y * c2->radius) + (c2->y * c1->radius)) / (c1->radius + c2->radius);
		return TRUE;
	}
	return FALSE;
}
#endif